@@ -174,6 +174,7 @@ class TabSqliteDb:
             self.db.execute('PRAGMA temp_store = MEMORY;')
             self.db.execute('PRAGMA journal_size_limit = 1000000;')
             self.db.execute('PRAGMA synchronous = NORMAL;')
+            self.db.execute('PRAGMA busy_timeout = 5000;')
         except Exception as error: # pylint: disable=broad-except:
             LOGGER.exception(
                 'Error while initializing database: %s: %s',
@@ -353,7 +354,7 @@ class TabSqliteDb:
                             os.rename(user_db+'-wal', user_db+'-wal'+timestamp)
                         LOGGER.debug(
                             'Creating a new, empty database "%s".', user_db)
-                        self.init_user_db(user_db)
+                        TabSqliteDb._init_user_db(user_db)
                         LOGGER.debug(
                             'If user phrases were successfully recovered from '
                             'the old, '
@@ -362,16 +363,17 @@ class TabSqliteDb:
                     else:
                         LOGGER.debug(
                             'Compatible database %s found.', user_db)
-                except:
+                except Exception as error: # pylint: disable=broad-except
                     LOGGER.exception(
-                        'Unexpected error trying to find user database')
+                        'Unexpected error trying to find user database: %s: %s',
+                        error.__class__.__name__, error)
 
         # open user phrase database
         try:
             LOGGER.debug(
                 'Connect to the database %s.', user_db)
-            self.db.executescript('''
-                ATTACH DATABASE "%s" AS user_db;
+            self.db.executescript(f'''
+                ATTACH DATABASE "{user_db}" AS user_db;
                 PRAGMA user_db.encoding = "UTF-8";
                 PRAGMA user_db.case_sensitive_like = true;
                 PRAGMA user_db.page_size = 4096;
@@ -380,9 +382,12 @@ class TabSqliteDb:
                 PRAGMA user_db.journal_mode = WAL;
                 PRAGMA user_db.journal_size_limit = 1000000;
                 PRAGMA user_db.synchronous = NORMAL;
-            ''' % user_db)
-        except:
-            LOGGER.debug('Could not open the database %s.', user_db)
+                PRAGMA busy_timeout = 5000;
+            ''')
+        except Exception as error:  # pylint: disable=broad-except
+            LOGGER.exception(
+                'Could not open the database %s: %s: %s',
+                user_db, error.__class__.__name__, error)
             timestamp = time.strftime('-%Y-%m-%d_%H:%M:%S')
             LOGGER.debug('Renaming the incompatible database to "%s".',
                          user_db+timestamp)
@@ -393,9 +398,9 @@ class TabSqliteDb:
             if os.path.exists(user_db+'-wal'):
                 os.rename(user_db+'-wal', user_db+'-wal'+timestamp)
             LOGGER.debug('Creating a new, empty database "%s".', user_db)
-            self.init_user_db(user_db)
-            self.db.executescript('''
-                ATTACH DATABASE "%s" AS user_db;
+            TabSqliteDb._init_user_db(user_db)
+            self.db.executescript(f'''
+                ATTACH DATABASE "{user_db}" AS user_db;
                 PRAGMA user_db.encoding = "UTF-8";
                 PRAGMA user_db.case_sensitive_like = true;
                 PRAGMA user_db.page_size = 4096;
@@ -404,7 +409,8 @@ class TabSqliteDb:
                 PRAGMA user_db.journal_mode = WAL;
                 PRAGMA user_db.journal_size_limit = 1000000;
                 PRAGMA user_db.synchronous = NORMAL;
-            ''' % user_db)
+                PRAGMA busy_timeout = 5000;
+            ''')
         self.create_tables("user_db")
         if self.old_phrases:
             sqlargs_old_phrases: List[Dict[str, Union[str, int]]] = []
@@ -420,8 +426,9 @@ class TabSqliteDb:
             '''
             try:
                 self.db.executemany(sqlstr, sqlargs_old_phrases)
-            except:
-                LOGGER.exception('Error inserting old phrases')
+            except sqlite3.Error as error:
+                LOGGER.exception('Error inserting old phrases: %s: %s',
+                                  error.__class__.__name__, error)
             self.db.commit()
             self.db.execute('PRAGMA wal_checkpoint;')
 
@@ -443,10 +450,10 @@ class TabSqliteDb:
                 tabkeys, phrase, user_freq, database)
         if not tabkeys or not phrase:
             return
-        sqlstr = '''
-        UPDATE %s.phrases SET user_freq = :user_freq
+        sqlstr = f'''
+        UPDATE {database}.phrases SET user_freq = :user_freq
         WHERE tabkeys = :tabkeys AND phrase = :phrase
-        ;''' % database
+        ;'''
         sqlargs = {'user_freq': user_freq,
                    'tabkeys': tabkeys,
                    'phrase': phrase}
@@ -455,8 +462,10 @@ class TabSqliteDb:
             if commit:
                 self.db.commit()
             self.invalidate_phrases_cache(tabkeys)
-        except:
-            LOGGER.exception('Unexpected error updating phrase in user_db.')
+        except sqlite3.Error as error:
+            LOGGER.exception(
+                'Unexpected error updating phrase in user_db: %s: %s',
+                error.__class__.__name__, error)
 
     def sync_usrdb(self) -> None:
         '''
@@ -496,7 +505,8 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug('load_phrases_cache()')
         try:
-            self._phrases_cache = json.load(open(self.cache_path))
+            with open(self.cache_path, encoding='utf-8') as f:
+                self._phrases_cache = json.load(f)
             snum = self._phrases_cache.get('serial_number')
             if not snum or (snum != self._snum):
                 self._phrases_cache = {}
@@ -508,8 +518,10 @@ class TabSqliteDb:
             if DEBUG_LEVEL > 1:
                 LOGGER.debug(
                     'Permission error reading %s', self.cache_path)
-        except:
-            LOGGER.debug('Unknown error reading %s', self.cache_path)
+        except Exception as error: # pylint: disable=broad-except
+            LOGGER.exception(
+                'Unknown error reading %s: %s: %s',
+                self.cache_path, error.__class__.__name__, error)
 
     def save_phrases_cache(self) -> None:
         '''
@@ -520,12 +532,15 @@ class TabSqliteDb:
         try:
             self._phrases_cache['serial_number'] = self._snum
             _cache_path = self.cache_path + '.tmp'
-            # The system may be break stores on rebooting, so
+            # The system may be break during rebooting, so
             # dump to temporary file and then replace it.
-            json.dump(self._phrases_cache, open(_cache_path, 'w'))
+            with open(_cache_path, 'w', encoding='utf-8') as f:
+                json.dump(self._phrases_cache, f)
             os.replace(_cache_path, self.cache_path)
-        except:
-            LOGGER.exception('Unexpected error in save_phrases_cache().')
+        except Exception as error: # pylint: disable=broad-except
+            LOGGER.exception(
+                'Unexpected error in save_phrases_cache(): %s: %s',
+                error.__class__.__name__, error)
 
     def is_chinese(self) -> bool:
         '''
@@ -591,27 +606,27 @@ class TabSqliteDb:
     def create_tables(self, database: str) -> None:
         '''Create tables that contain all phrase'''
         if database == 'main':
-            sqlstr = '''
-            CREATE TABLE IF NOT EXISTS %s.goucima
+            sqlstr = f'''
+            CREATE TABLE IF NOT EXISTS {database}.goucima
             (zi TEXT PRIMARY KEY, goucima TEXT);
-            ''' % database
+            '''
             self.db.execute(sqlstr)
-            sqlstr = '''
-            CREATE TABLE IF NOT EXISTS %s.pinyin
+            sqlstr = f'''
+            CREATE TABLE IF NOT EXISTS {database}.pinyin
             (pinyin TEXT, zi TEXT, freq INTEGER);
-            ''' % database
+            '''
             self.db.execute(sqlstr)
-            sqlstr = '''
-            CREATE TABLE IF NOT EXISTS %s.suggestion
+            sqlstr = f'''
+            CREATE TABLE IF NOT EXISTS {database}.suggestion
             (phrase TEXT, freq INTEGER);
-            ''' %database
+            '''
             self.db.execute(sqlstr)
 
-        sqlstr = '''
-        CREATE TABLE IF NOT EXISTS %s.phrases
+        sqlstr = f'''
+        CREATE TABLE IF NOT EXISTS {database}.phrases
         (id INTEGER PRIMARY KEY, tabkeys TEXT, phrase TEXT,
         freq INTEGER, user_freq INTEGER);
-        ''' % database
+        '''
         self.db.execute(sqlstr)
         self.db.commit()
 
@@ -677,7 +692,7 @@ class TabSqliteDb:
                         rules['above'] = int(res.group(2))
                     _cms = res.group(3).split('+')
                     if len(_cms) > self._mlen:
-                        print('rule: "%s" over max key length' %rule)
+                        print(f'rule: "{rule}" over max key length')
                         break
                     for _cm in _cms:
                         cm_res = patt_p.match(_cm)
@@ -686,9 +701,11 @@ class TabSqliteDb:
                                         int(cm_res.group(2))))
                     rules[int(res.group(2))] = cms
                 else:
-                    print('not a legal rule: "%s"' %rule)
-        except Exception:
-            LOGGER.exception('Unexpected error in get_rules().')
+                    print(f'not a legal rule: "{rule}"')
+        except Exception as error: # pylint: disable=broad-except
+            LOGGER.exception(
+                'Unexpected error in get_rules(): %s: %s',
+                error.__class__.__name__, error)
         return rules
 
     def get_possible_tabkeys_lengths(self) -> List[int]:
@@ -758,11 +775,11 @@ class TabSqliteDb:
         '''
         if DEBUG_LEVEL > 1:
             LOGGER.debug('len(phrases)=%s', len(list(phrases)))
-        insert_sqlstr = '''
+        insert_sqlstr = f'''
         INSERT INTO {database}.phrases
         (tabkeys, phrase, freq, user_freq)
         VALUES (:tabkeys, :phrase, :freq, :user_freq);
-        '''.format(database=database)
+        '''
         insert_sqlargs = []
         for (tabkeys, phrase, freq, user_freq) in phrases:
             insert_sqlargs.append({
@@ -793,10 +810,10 @@ class TabSqliteDb:
                 tabkeys, phrase, freq, user_freq)
         if not tabkeys or not phrase:
             return
-        select_sqlstr = '''
+        select_sqlstr = f'''
         SELECT * FROM {database}.phrases
         WHERE tabkeys = :tabkeys AND phrase = :phrase;
-        '''.format(database=database)
+        '''
         select_sqlargs = {'tabkeys': tabkeys, 'phrase': phrase}
         results = self.db.execute(select_sqlstr, select_sqlargs).fetchall()
         if results:
@@ -809,11 +826,11 @@ class TabSqliteDb:
                     select_sqlstr, select_sqlargs, results)
             return
 
-        insert_sqlstr = '''
+        insert_sqlstr = f'''
         INSERT INTO {database}.phrases
         (tabkeys, phrase, freq, user_freq)
         VALUES (:tabkeys, :phrase, :freq, :user_freq);
-        '''.format(database=database)
+        '''
         insert_sqlargs = {
             'tabkeys': tabkeys,
             'phrase': phrase,
@@ -828,8 +845,10 @@ class TabSqliteDb:
             if commit:
                 self.db.commit()
             self.invalidate_phrases_cache(tabkeys)
-        except:
-            LOGGER.exception('Unexpected error in add_phrase()')
+        except Exception as error: # pylint: disable=broad-except
+            LOGGER.exception(
+                'Unexpected error in add_phrase(): %s: %s',
+                error.__class__.__name__, error)
 
     def add_goucima(self, goucimas: Iterable[Tuple[str, str]]) -> None:
         '''Add goucima into database, goucimas is iterable object
@@ -846,8 +865,10 @@ class TabSqliteDb:
             self.db.executemany(sqlstr, sqlargs)
             self.db.commit()
             self.db.execute('PRAGMA wal_checkpoint;')
-        except:
-            LOGGER.exception('Unexpected error in add_goucima().')
+        except sqlite3.Error as error:
+            LOGGER.exception(
+                'Unexpected error in add_goucima(): %s: %s',
+                error.__class__.__name__, error)
 
     def add_pinyin(
             self,
@@ -856,9 +877,10 @@ class TabSqliteDb:
         '''Add pinyin to database, pinyins is a iterable object
         Like: [(zi,pinyin, freq), (zi, pinyin, freq), ...]
         '''
-        sqlstr = '''
-        INSERT INTO %s.pinyin (pinyin, zi, freq) VALUES (:pinyin, :zi, :freq);
-        ''' % database
+        sqlstr = f'''
+        INSERT INTO {database}.pinyin (pinyin, zi, freq)
+        VALUES (:pinyin, :zi, :freq);
+        '''
         count = 0
         for pinyin, zi, freq in pinyins:
             count += 1
@@ -871,11 +893,13 @@ class TabSqliteDb:
             try:
                 self.db.execute(
                     sqlstr, {'pinyin': pinyin, 'zi': zi, 'freq': freq})
-            except Exception:
+            except sqlite3.Error as error:
                 LOGGER.exception(
                     'Error when inserting into pinyin table. '
-                    'count=%s pinyin=%s zi=%s freq=%s',
-                    count, pinyin, zi, freq)
+                    'count=%s pinyin=%s zi=%s freq=%s: '
+                    '%s: %s',
+                    count, pinyin, zi, freq,
+                    error.__class__.__name__, error)
         self.db.commit()
 
     def add_suggestion(
@@ -885,20 +909,21 @@ class TabSqliteDb:
         '''Add suggestion phrase to database, suggestions is a iterable object
         Like: [(phrase, freq), (phrase, freq), ...]
         '''
-        sqlstr = '''
-        INSERT INTO %s.suggestion (phrase, freq) VALUES (:phrase, :freq);
-        ''' % database
+        sqlstr = f'''
+        INSERT INTO {database}.suggestion (phrase, freq) VALUES (:phrase, :freq);
+        '''
         count = 0
         for phrase, freq in suggestions:
             count += 1
             try:
                 self.db.execute(
                     sqlstr, {'phrase': phrase, 'freq': freq})
-            except Exception:
+            except sqlite3.Error as error:
                 LOGGER.exception(
                     'Error when inserting into suggestion table. '
-                    'count=%s phrase=%s freq=%s',
-                    count, phrase, freq)
+                    'count=%s phrase=%s freq=%s: %s: %s',
+                    count, phrase, freq,
+                    error.__class__.__name__, error)
         self.db.commit()
 
     def optimize_database(self) -> None:
@@ -929,7 +954,8 @@ class TabSqliteDb:
         self.db.executescript("VACUUM;")
         self.db.commit()
 
-    def drop_indexes(self, _database: str) -> None:
+    def drop_indexes( # pylint: disable=no-self-use
+            self, _database: str) -> None:
         '''Drop the indexes in the database to reduce its size
 
         We do not use any indexes at the moment, therefore this
@@ -938,7 +964,8 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug('drop_indexes()')
 
-    def create_indexes(self, _database: str, _commit: bool = True) -> None:
+    def create_indexes( # pylint: disable=no-self-use
+            self, _database: str, _commit: bool = True) -> None:
         '''Create indexes for the database.
 
         We do not use any indexes at the moment, therefore
@@ -953,12 +980,14 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug('create_indexes()')
 
-    def big5_code(self, phrase: str) -> bytes:
+    @staticmethod
+    def _big5_code(phrase: str) -> bytes:
         '''
         Encode a string in Big5 or, if that is not possible,
         return something higher than any Big5 code.
 
         :param phrase: String to be encoded in Big5 encoding
+        :returns: Big5-encoded bytes or b'\xff\xff' if unencodable
         '''
         try:
             big5 = phrase.encode('Big5')
@@ -985,13 +1014,14 @@ class TabSqliteDb:
         if engine_name in [
                 'cangjie3', 'cangjie5', 'cangjie-big',
                 'quick-classic', 'quick3', 'quick5']:
-            code_point_function = self.big5_code
+            code_point_function = TabSqliteDb._big5_code
         if self.is_db_chinese:
-            pinyin_exact_match_function: Callable[[str], int] = lambda x: (
-                - int(typed_tabkeys == x[:-1] and  x[-1] in '!@#$%')
-            )
+            def pinyin_exact_match_function(x: str) -> int:
+                return -int(typed_tabkeys == x[:-1] and x[-1] in '!@#$%')
         else:
-            pinyin_exact_match_function = lambda x: (1)
+            def pinyin_exact_match_function( # pylint: disable=unused-argument
+                    x: str) -> int:
+                return 1
         if chinese_mode in (2, 3) and self.is_db_chinese:
             if chinese_mode == 2:
                 bitmask = 1 << 0 # used in simplified Chinese
@@ -1055,7 +1085,7 @@ class TabSqliteDb:
             one_char_condition = ' AND length(phrase)=1 '
 
         if self.user_can_define_phrase or dynamic_adjust:
-            sqlstr = '''
+            sqlstr = f'''
             SELECT tabkeys, phrase, freq, user_freq FROM
             (
                 SELECT tabkeys, phrase, freq, user_freq FROM main.phrases
@@ -1064,12 +1094,12 @@ class TabSqliteDb:
                 SELECT tabkeys, phrase, freq, user_freq FROM user_db.phrases
                 WHERE tabkeys LIKE :tabkeys ESCAPE :escapechar {one_char_condition}
             )
-            '''.format(one_char_condition=one_char_condition)
+            '''
         else:
-            sqlstr = '''
+            sqlstr = f'''
             SELECT tabkeys, phrase, freq, user_freq FROM main.phrases
             WHERE tabkeys LIKE :tabkeys ESCAPE :escapechar {one_char_condition}
-            '''.format(one_char_condition=one_char_condition)
+            '''
         escapechar = '☺'
         for char in '!@#':
             if char not in [single_wildcard_char, multi_wildcard_char]:
@@ -1086,9 +1116,9 @@ class TabSqliteDb:
                 single_wildcard_char, '_')
         if multi_wildcard_char:
             tabkeys_for_like = tabkeys_for_like.replace(
-                multi_wildcard_char, '%%')
+                multi_wildcard_char, '%')
         if auto_wildcard:
-            tabkeys_for_like += '%%'
+            tabkeys_for_like += '%'
         sqlargs = {'tabkeys': tabkeys_for_like, 'escapechar': escapechar}
         if DEBUG_LEVEL > 1:
             LOGGER.debug('sqlstr=%s sqlargs=%s', sqlstr, repr(sqlargs))
@@ -1212,7 +1242,7 @@ class TabSqliteDb:
         if engine_name in [
                 'cangjie3', 'cangjie5', 'cangjie-big',
                 'quick-classic', 'quick3', 'quick5']:
-            code_point_function = self.big5_code
+            code_point_function = TabSqliteDb._big5_code
 
         return sorted(candidates,
                       key=lambda x: (
@@ -1235,19 +1265,23 @@ class TabSqliteDb:
         try:
             sqlstring = (
                 'CREATE TABLE IF NOT EXISTS user_db.desc '
-                + '(name PRIMARY KEY, value);')
+                '(name PRIMARY KEY, value);')
             self.db.executescript(sqlstring)
             sqlstring = 'INSERT OR IGNORE INTO user_db.desc  VALUES (?, ?);'
-            self.db.execute(sqlstring, ('version', DATABASE_VERSION))
+            self.db.execute(
+                sqlstring, ('version', DATABASE_VERSION))
             sqlstring = (
-                'INSERT OR IGNORE INTO user_db.desc  '
-                + 'VALUES (?, DATETIME("now", "localtime"));')
-            self.db.execute(sqlstring, ("create-time", ))
+                "INSERT OR IGNORE INTO user_db.desc  "
+                "VALUES ('create-time', DATETIME('now', 'localtime'));")
+            self.db.execute(sqlstring)
             self.db.commit()
-        except:
-            LOGGER.exception('Unexpected error in generate_userdb_desc().')
+        except Exception as error: # pylint: disable=broad-except
+            LOGGER.exception(
+                'Unexpected error adding description to user_db: %s: %s',
+                 error.__class__.__name__, error)
 
-    def init_user_db(self, db_file: str) -> None:
+    @staticmethod
+    def _init_user_db(db_file: str) -> None:
         '''
         Initialize the user database unless it is an in-memory database
 
@@ -1268,6 +1302,7 @@ class TabSqliteDb:
                 PRAGMA journal_mode = WAL;
                 PRAGMA journal_size_limit = 1000000;
                 PRAGMA synchronous = NORMAL;
+                PRAGMA busy_timeout = 5000;
             ''')
             db.commit()
 
@@ -1427,10 +1462,11 @@ class TabSqliteDb:
                 zi -= 1
             if ma > 0:
                 ma -= 1
-            tabkey = self.get_goucima(phrase[zi])[ma]
-            if not tabkey:
+            goucima = self.get_goucima(phrase[zi])
+            if len(goucima) < ma + 1:
+                LOGGER.error('goucima=%r too short no index ma=%s', goucima, ma)
                 return ''
-            tabkeys += tabkey
+            tabkeys += goucima[ma]
         if DEBUG_LEVEL > 1:
             LOGGER.debug('tabkeys=%s', tabkeys)
         return tabkeys
@@ -1578,15 +1614,15 @@ class TabSqliteDb:
         if not phrase:
             return
         if tabkeys:
-            delete_sqlstr = '''
+            delete_sqlstr = f'''
             DELETE FROM {database}.phrases
             WHERE tabkeys = :tabkeys AND phrase = :phrase;
-            '''.format(database=database)
+            '''
         else:
-            delete_sqlstr = '''
+            delete_sqlstr = f'''
             DELETE FROM {database}.phrases
             WHERE phrase = :phrase;
-            '''.format(database=database)
+            '''
         delete_sqlargs = {'tabkeys': tabkeys, 'phrase': phrase}
         self.db.execute(delete_sqlstr, delete_sqlargs)
         if commit:
